/*---------------------------------------------------------------*/
/* 8-pin SD audio player R0.04                     (C)ChaN, 2010 */
/*---------------------------------------------------------------*/
#define F_CPU 16000000
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "diskio.h"
#include "pff.h"
#include "fix_fft.h"

#define FCC(c1,c2,c3,c4)	(((DWORD)c4<<24)+((DWORD)c3<<16)+((WORD)c2<<8)+(BYTE)c1)	/* FourCC */

#define FOSC 16000000
#define BAUD 9600
#define MYUBRR (FOSC/16/BAUD - 1)
#define LOG_NUM_SAMPLES 5
#define NUM_SAMPLES (1 << LOG_NUM_SAMPLES)
#define NUM_LEVELS 8

/*---------------------------------------------------------*/
/* Work Area                                               */
/*---------------------------------------------------------*/

volatile BYTE FifoRi, FifoWi, FifoCt;	/* FIFO controls */

unsigned char Buff1[256];		/* Wave output FIFO */
unsigned char Buff2[256];		/* Wave output FIFO */
unsigned char re[32];
unsigned char im[32];
unsigned char lit[8];
uint8_t rewind;

FATFS Fs;			/* File system object */
DIR Dir;			/* Directory object */
FILINFO Fno;		/* File information */

WORD rb;			/* Return value. Put this here to avoid bugs of avr-gcc */


/*---------------------------------------------------------*/

int uart_putchar(char c, FILE *unused)      
{
        if (c == '\n')
                uart_putchar('\r', 0);
        loop_until_bit_is_set(UCSRA, UDRE);
        UDR = c;
        return 0;
}

// Functie ce trateaza primirea unui caracter pe seriala
int uart_getchar(FILE* f)
{
        char c;

        loop_until_bit_is_set(UCSRA, RXC);
        c = UDR;
        if(c=='\r')
                c='\n';
        uart_putchar(c,NULL); 
        return c;
}


static
DWORD load_header (void)	/* 0:Invalid format, 1:I/O error, >1:Number of samples */
{
        DWORD sz;


        if (pf_read(Buff1, 12, &rb)) return 1;	/* Load file header (12 bytes) */

        if (rb != 12 || LD_DWORD(Buff1+8) != FCC('W','A','V','E')) return 0;

        for (;;) {
                pf_read(Buff1, 8, &rb);			/* Get Chunk ID and size */
                if (rb != 8) return 0;
                sz = LD_DWORD(&Buff1[4]);		/* Chunk size */

                switch (LD_DWORD(&Buff1[0])) {	/* FCC */
                case FCC('f','m','t',' ') :					/* 'fmt ' chunk */
                        if (sz > 100 || sz < 16) return 0;		/* Check chunk size */
                        pf_read(Buff1, sz, &rb);					/* Get content */
                        if (rb != sz) return 0;
                        if (Buff1[0] != 1) return 0;				/* Check coding type (1) */
                        if (Buff1[2] != 1 && Buff1[2] != 2) 		/* Check channels (1/2) */
                                return 0;

                        if (Buff1[14] != 8 && Buff1[14] != 16)	/* Check resolution (8/16) */
                                return 0;

                        OCR0 = (BYTE)(F_CPU/8/LD_WORD(&Buff1[4]))-1;	/* Sampling freq */

                        break;

                case FCC('d','a','t','a') :				/* 'data' chunk (start to play) */
                        return sz;

                case FCC('L','I','S','T') :				/* 'LIST' chunk (skip) */
                case FCC('f','a','c','t') :				/* 'fact' chunk (skip) */
                        pf_lseek(Fs.fptr + sz);
                        break;

                default :								/* Unknown chunk (error) */
                        return 0;
                }
        }

        return 0;
}




volatile uint8_t read1 = 1;

volatile uint16_t counter=0;

volatile uint8_t sync = 0;

volatile uint8_t vu=0;

ISR(TIMER0_COMP_vect)
{


        if(sync == 1)
        {

                if(read1 == 1)
                {
                        OCR1A = (unsigned char)Buff1[counter];
                        OCR1B = (unsigned char)Buff1[counter++];

                        if(counter == 256)
                        {
                                read1 = 0;
                                counter = 0;
                        }
                }
                else if(read1 == 0)
                {
                        OCR1A = (unsigned char)Buff2[counter];
                        OCR1B = (unsigned char)Buff2[counter++];
                        if(counter == 256)
                        {
                                read1 = 1;
                                counter = 0;
                        }
                }
        }
}

ISR(TIMER2_COMP_vect)
{
        int i, j;
        for (i = 0; i < NUM_LEVELS; ++i)
                for (j = 0; j < NUM_LEVELS; ++j) {
                        if (lit[i] & _BV(j)) {
                                PORTA |= _BV(i);
                                PORTC &= ~_BV(j);
                                _delay_ms(0.01);
                        } else {
                                PORTA &= ~_BV(i);
                                PORTC |= _BV(j);
                        }
                }
}

void Timer0_init(void)
{
        TCCR0 = _BV(WGM01)|_BV(CS01);//CTC, FCPU/8
        TIMSK = _BV(OCIE0);
}

void Timer2_init(void)
{
        TCCR2 =  _BV(WGM21) | _BV(CS22) | _BV(CS21) | _BV(CS20);
        TIMSK = _BV(OCIE2);
}

void Timer1_init(void)
{

        TCCR1A = _BV(COM1A1) | _BV(COM1A0) | _BV(COM1B1) | _BV(COM1B0) |
                _BV(WGM10);
        TCCR1B = _BV(WGM12) | _BV(CS10);
}


static UINT play (const char *fn)
{
        DWORD sz;
        FRESULT res;
        unsigned char read2;


        uint16_t size = 256;
        unsigned char* buf;
        int i, j, k;
        int old_i = 0, old_j = 0;
        uint32_t l = 0;
        PORTA = 0;
        PORTC = 0xff;

        if ((res = pf_open(fn)) == FR_OK) {
                Timer0_init();
                Timer1_init();
                /*Timer2_init();*/

                sz = load_header();			/* Load file header */
                if (sz < 256) return (UINT)sz;


                FifoCt = 0; FifoRi = 0; FifoWi = 0;		/* Reset FIFO */
                pf_read(Buff1, 512 - (Fs.fptr % 512), &rb);	/* Snip sector unaligned part */
                sz -= rb;


                res = pf_read(Buff1, size, &rb);
                if (res != FR_OK) { while (FifoCt) ; return res; }

                sync = 1;

                do {

                        read2 = read1;

                        if(read1 == 1) {
                                res = pf_read(Buff2, size, &rb);
                                buf = Buff2;
                        } else {
                                res = pf_read(Buff1, size, &rb);
                                buf = Buff1;
                        }

                        if (l % 10 == 0) {
                                /*for (k = 0; k < NUM_SAMPLES; ++k) {*/
                                        /*re[k] = buf[k];*/
                                        /*im[k] = 0;*/
                                /*}*/
                                /*fix_fft(re, im, LOG_NUM_SAMPLES, 0);*/
                                /*for (k = 0; k < NUM_SAMPLES;) {*/
                                        /*int sum = 0;*/
                                        /*for (i = k; k < i + NUM_SAMPLES / NUM_LEVELS; ++k) {*/
                                                /*sum += sqrt(re[k] * re[k] + im[k] **/
                                                                /*im[k]);*/
                                        /*}*/
                                        /*i = k / NUM_SAMPLES / NUM_LEVELS;*/
                                        /*for (j = 0; j < NUM_LEVELS; ++j) {*/
                                                /*if (sum > 0) {*/
                                                        /*lit[i] |= _BV(j);*/
                                                        /*sum -= 255;*/
                                                /*} else {*/
                                                        /*lit[i] &= ~_BV(j);*/
                                                /*}*/
                                        /*}*/
                                /*}*/
                                i = rand() % NUM_LEVELS;
                                j = rand () % NUM_LEVELS;
                                PORTA &= ~_BV(old_i);
                                PORTC |= _BV(old_j);
                                PORTA |= _BV(i);
                                PORTC &= ~_BV(j);
                                old_i = i;
                                old_j = j;
                        }
                        ++l;

                        if (res != FR_OK) 
                                break;

                        while (read2 == read1);

                        if (!(PINB & _BV(PB2))) {
                                sync = 0;
                                _delay_ms(500); 
                                break;
                        }
                        if (!(PINB & _BV(PB3))) {
                                sync = 0;
                                rewind = 1;
                                _delay_ms(500); 
                                break;
                        }
                } while (rb == size); /* Repeat until all data read */
                sync = 0;
        }

        while (FifoCt) ; /* Wait for FIFO empty */
        return res;
}

/*-----------------------------------------------------------------------*/
/* Main                                                                  */
void USART_init(void)
{
        UBRRL = (unsigned char)MYUBRR;
        UBRRL = 103;
        UCSRB = _BV(RXEN) | _BV(TXEN); // deschidere trans; recv

        UCSRC = _BV(URSEL) |  (3 << UCSZ0); // formatul pentru un cadru: 8 biti data; 2 biti stop; sel UCSRC
}

int main (void)
{

        USART_init();	
        DDRD |= (1<<PD5)|(1<<PD6) | _BV(4);
        DDRC = 0xFF;
        DDRB &= ~(_BV(PB3) | _BV(PB2));
        PORTB |= _BV(PB3) | _BV(PB2);
        PORTD &= ~((1<<PD5)|(1<<PD6) | _BV(PD4));

        sei();

        for (;;) {
                _delay_ms(1000);					/* Delay 1000ms */
                if (pf_mount(&Fs)) {
                        continue; /* Initialize FS */
                }
                for (;;) {
                        if (pf_opendir(&Dir, "")) break;	/* Open sound file directory (root dir) */
                        while (!pf_readdir(&Dir, &Fno) && Fno.fname[0]) {	/* Play all wav files in the dir */
                                if (!(Fno.fattrib & (AM_DIR|AM_HID)) && strstr(Fno.fname, ".WAV")) {
                                        do {
                                                rewind = 0;
                                                play(Fno.fname);
                                        } while (rewind);
                                }
                        }
                }
        }
}

